from neowg.key_pair import KeyPair


class WgClientConfig:
    _server_host: str
    _server_pubkey: str
    _ip: str
    _keys: KeyPair
    _used: bool

    def __init__(
            self, ip: str, server_host: str, server_pubkey: str, keys: KeyPair | None = None, used: bool = False
    ) -> None:
        """Создает пользовательский конфиг с указанной связкой ключей или генерирует новую."""
        self._ip = ip
        self._keys = keys or KeyPair()
        self._server_host = server_host
        self._server_pubkey = server_pubkey
        self._used = used

    def dump(self) -> str:
        """Записывает пользовательский конфиг по указанному пути."""

        file_content = """[Interface]
Address = {client_ip}/32
PrivateKey = {private_key}
DNS = 8.8.8.8

[Peer]
PublicKey = {server_pubkey}
Endpoint = {server_host}
AllowedIPs = 0.0.0.0/0"""
        file_content = file_content.format(
            private_key=self._keys.private_key,
            server_pubkey=self._server_pubkey,
            server_host=self._server_host,
            client_ip=self._ip,
        )

        return file_content

    def update_keys(self) -> KeyPair:
        """Обновляет ключи в конфиге и помечает его как свободный"""
        new_keys = KeyPair()
        self._keys = new_keys
        self._used = False
        return new_keys

    @property
    def pubkey(self) -> str:
        return self._keys.pubkey

    @property
    def ip(self) -> str:
        return self._ip

    @property
    def private_key(self) -> str:
        return self._keys.private_key

    @property
    def used(self) -> bool:
        return self._used

    def mark_used(self) -> None:
        self._used = True

    def __str__(self) -> str:
        return (f"WgClientConfig(server_host={self._server_host}, server_pubkey={self._server_pubkey}), "
                f"ip={self._ip}, keys={self._keys}, used={self._used})")



import subprocess

__all__ = [
    "KeyPair",
]


class KeyPair:
    _pubkey: str
    _private_key: str

    def __init__(self, private_key: str | None = None, *, pubkey: str | None = None) -> None:
        if private_key is not None and pubkey is not None:
            self._pubkey = pubkey
            self._private_key = private_key
        elif private_key is not None and pubkey is None:
            self._private_key = private_key
            self._pubkey = self._generate_pubkey(self._private_key)
        elif private_key is None and pubkey is None:
            self._pubkey, self._private_key = self._generate_key_pair()
        else:
            raise ValueError("Invalid arguments")

    def _generate_key_pair(self) -> tuple[str, str]:
        private_key = (subprocess
                       .run(["wg", "genkey"], capture_output=True, text=True)
                       .stdout
                       .replace("\n", "")
                       )
        public_key = self._generate_pubkey(private_key)
        return public_key, private_key

    @staticmethod
    def _generate_pubkey(private_key: str) -> str:
        return (subprocess
                .run(["wg", "pubkey"], capture_output=True, text=True, input=private_key)
                .stdout
                .replace("\n", "")
                )

    @property
    def pubkey(self) -> str:
        return self._pubkey

    @property
    def private_key(self) -> str:
        return self._private_key

    def __str__(self) -> str:
        return f'KeyPair(pubkey="{self._pubkey}", private_key="{self._private_key}")'



from .key_pair import KeyPair
from .wg_client_config import WgClientConfig
from .wg_server_config import WgServerConfig
from .wg_service import WgService

__all__ = [
    "KeyPair",
    "WgClientConfig",
    "WgServerConfig",
    "WgService",
]



from pathlib import Path
from typing import Self

from neowg.wg_client_config import WgClientConfig
from neowg.key_pair import KeyPair
from neowg.config import WIREGUARD_PORT
from neowg.utils import get_server_ip, get_server_net_interface


class WgServerConfig:
    _keys: KeyPair
    _clients: list[WgClientConfig]
    _server_ip: str
    _net_adapter: str

    @classmethod
    def new(cls: type[Self], clients_count: int = 100) -> Self:
        """Создание нового конфига."""
        config = cls()
        config._server_ip = get_server_ip()
        config._keys = KeyPair()
        config._net_adapter = get_server_net_interface()

        config._clients = []

        last_octec = 0
        pre_last_octec = 0
        i = 0
        octecs = []

        while i < clients_count:
            if i % 255 == 0 and i != 0:
                last_octec = 0
                pre_last_octec += 1
            else:
                last_octec += 1

            i += 1
            octecs.append((pre_last_octec, last_octec))


        for pre_last_octec, last_octec in octecs:
            if not (pre_last_octec == 0 and last_octec in (0, 1)):
                config._clients.append(WgClientConfig(
                    ip=f"10.0.{pre_last_octec}.{last_octec}",
                    server_host=f"{config._server_ip}:{WIREGUARD_PORT}",
                    server_pubkey=config._keys.pubkey,
                ))

        return config

    @classmethod
    def from_file(cls: type[Self], path: Path | str) -> Self:
        """Чтение wireguard конфига из файла."""
        path = Path(path)
        config = cls()

        with path.open() as f:
            interface, *peers = f.read().split("\n\n")

        config._keys = cls._parse_server_keys(interface)
        config._server_ip = cls._parse_server_ip(interface)
        config._clients = cls._parse_clients(peers, config._keys.pubkey, f"{config._server_ip}:{WIREGUARD_PORT}")
        config._net_adapter = cls._parse_net_adapter(interface)

        return config

    def dump(self, path: Path | str) -> None:
        """Запись wireguard конфига в файл по указанному пути."""

        path = Path(path)

        file_content = """[Interface]
Address = 10.0.0.1/24
PostUp = iptables -A FORWARD -i %i -j ACCEPT; iptables -A FORWARD -o %i -j ACCEPT; iptables -t nat -A POSTROUTING -o {net_adapter} -j MASQUERADE
PostDown = iptables -D FORWARD -i %i -j ACCEPT; iptables -D FORWARD -o %i -j ACCEPT; iptables -t nat -D POSTROUTING -o {net_adapter} -j MASQUERADE
ListenPort = {server_port}
PrivateKey = {private_key}
# PublicKey = {pubkey}
# ServerIP = {server_ip}
"""
        file_content = file_content.format(
            server_port=51280,
            private_key=self._keys.private_key,
            pubkey=self._keys.pubkey,
            server_ip=self._server_ip,
            net_adapter=self._net_adapter,
        )

        for client in self._clients:
            client_content = """
[Peer]
AllowedIPs = {ip}/32
# PrivateKey = {private_key}
PublicKey = {pubkey}
# Used = {used}
"""
            file_content += client_content.format(
                pubkey=client.pubkey, ip=client.ip, private_key=client.private_key, used=client.used
            )

        with path.open("w", encoding="utf-8") as f:
            f.write(file_content)

    def update_keys(self, ip: str) -> KeyPair:
        """Обновление пары ключей у указанного ip. Вернет новую пару ключей."""

        for client in self._clients:
            if client.ip == ip:
                return client.update_keys()

    @staticmethod
    def _parse_server_keys(interface: str) -> KeyPair:
        *_, private_key, pubkey, server_ip = interface.splitlines()
        _, pubkey = pubkey.split(' = ')
        _, private_key = private_key.split(' = ')
        return KeyPair(pubkey=pubkey, private_key=private_key)

    @staticmethod
    def _parse_server_ip(interface: str) -> str:
        *_, server_ip = interface.splitlines()
        return server_ip.split(' = ')[-1]

    @staticmethod
    def _parse_clients(peers: list[str], server_pubkey: str, server_host: str) -> list[WgClientConfig]:
        clients: list[WgClientConfig] = []
        for peer in peers:
            *_, ip, private_key, pubkey, used = peer.splitlines()
            private_key = private_key.split(' = ')[-1]
            pubkey = pubkey.split(' = ')[-1]
            ip = ip.split(' = ')[-1].split('/')[0]
            used = used.split(' = ')[-1].split('/')[0]

            clients.append(WgClientConfig(
                ip=ip,
                server_host=server_host,
                server_pubkey=server_pubkey,
                keys=KeyPair(pubkey=pubkey, private_key=private_key),
                used=used == "True",
            ))

        return clients

    @staticmethod
    def _parse_net_adapter(interface: str) -> str:
        lines = interface.splitlines()
        return lines[2].split('POSTROUTING -o ')[-1].replace(' -j MASQUERADE', '')

    def __str__(self) -> str:
        return f'WgServerConfig(keys={self._keys}, server="{self._server_ip}", clients={self._clients})'

    @property
    def clients(self) -> list[WgClientConfig]:
        return self._clients.copy()

    def allocate_config(self) -> WgClientConfig | None:
        for i in range(len(self._clients)):
            if not self._clients[i].used:
                self._clients[i].mark_used()
                return self._clients[i]
        return None


import shutil
import subprocess

from neowg import WgServerConfig
from neowg.config import SERVER_CONFIG_PATH

__all__ = [
    "WgService",
]


class WgService:
    def start(self) -> bool:
        if not self._check_wireguard_exists():
            print("[WgService] WireGuard not found. Trying install wireguard.")
            if not self._install_wireguard():
                return False

        print("[WgService] WireGuard found.")
        if not self._check_config_exists():
            self._create_server_config()

        if not self._is_wireguard_running():
            print("[WgService] WireGuard is not running.")
            if not self._start_wireguard_service():
                return False

        return True

    def _check_wireguard_exists(self) -> bool:
        print("[WgService] Checking if WireGuard exists")
        return shutil.which("wg") is not None

    def _install_wireguard(self) -> bool:
        print("[WgService] Installing WireGuard")
        command = "apt install wireguard"
        result = subprocess.run(command, shell=True,capture_output=True)

        if result.returncode == 0:
            print("[WgService] WireGuard installed")
            return True

        print("[WgService] Installing WireGuard failed")
        print(f"[WgService] stderr='{result.stderr.decode().strip()}'")
        print(f"[WgService] stdout='{result.stdout.decode().strip()}'")
        return False

    def _check_config_exists(self) -> bool:
        print("[WgService] Checking if config file exists")
        return SERVER_CONFIG_PATH.exists()

    def _create_server_config(self) -> None:
        print("[WgService] Creating config file")
        config = WgServerConfig.new()
        config.dump(SERVER_CONFIG_PATH)
        print("[WgService] Config file created")

    def _is_wireguard_running(self) -> bool:
        print("[WgService] Check WireGuard Service up")
        command = "systemctl status wg-quick@wg0"
        result = subprocess.run(command, shell=True, capture_output=True)

        return result.returncode == 0

    def _start_wireguard_service(self) -> bool:
        print("[WgService] Starting WireGuard service")
        command = "systemctl start wg-quick@wg0"
        result = subprocess.run(command, shell=True, capture_output=True)

        if result.returncode == 0:
            print("[WgService] WireGuard Service started")
            return True

        print("[WgService] WireGuard Service start failed")
        print(f"[WgService] stderr='{result.stderr.decode().strip()}'")
        print(f"[WgService] stdout='{result.stdout.decode().strip()}'")
        return False
